"""Low-level protocol stuff."""

from __future__ import annotations
from asyncio import StreamReader
from enum import Enum
from functools import partial
from logging import getLogger
from random import randint
from typing import IO, NamedTuple


__all__ = ['LittleEndianSignedInt32', 'Type', 'Packet', 'random_request_id']


LOGGER = getLogger(__file__)
TERMINATOR = b'\x00\x00'


def random_request_id() -> LittleEndianSignedInt32:
    """Generates a random request ID."""

    return LittleEndianSignedInt32(randint(0, LittleEndianSignedInt32.MAX))


class LittleEndianSignedInt32(int):
    """A little-endian, signed int32."""

    MIN = -2_147_483_648
    MAX = 2_147_483_647

    def __init__(self, *_):
        """Checks the boundaries."""
        super().__init__()

        if not self.MIN <= self <= self.MAX:
            raise ValueError('Signed int32 out of bounds:', int(self))

    def __bytes__(self):
        """Returns the integer as signed little endian."""
        return self.to_bytes(4, 'little', signed=True)

    @classmethod
    async def aread(cls, reader: StreamReader) -> LittleEndianSignedInt32:
        """Reads the integer from an asynchronous file-like object."""
        return cls.from_bytes(await reader.read(4), 'little', signed=True)

    @classmethod
    def read(cls, file: IO) -> LittleEndianSignedInt32:
        """Reads the integer from a file-like object."""
        return cls.from_bytes(file.read(4), 'little', signed=True)


class Type(Enum):
    """RCON packet types."""

    SERVERDATA_AUTH = LittleEndianSignedInt32(3)
    SERVERDATA_AUTH_RESPONSE = LittleEndianSignedInt32(2)
    SERVERDATA_EXECCOMMAND = LittleEndianSignedInt32(2)
    SERVERDATA_RESPONSE_VALUE = LittleEndianSignedInt32(0)

    def __int__(self):
        """Returns the actual integer value."""
        return int(self.value)

    def __bytes__(self):
        """Returns the integer value as little endian."""
        return bytes(self.value)

    @classmethod
    async def aread(cls, reader: StreamReader) -> Type:
        """Reads the type from an asynchronous file-like object."""
        return cls(await LittleEndianSignedInt32.aread(reader))

    @classmethod
    def read(cls, file: IO) -> Type:
        """Reads the type from a file-like object."""
        return cls(LittleEndianSignedInt32.read(file))


class Packet(NamedTuple):
    """An RCON packet."""

    id: LittleEndianSignedInt32
    type: Type
    payload: bytes
    terminator: bytes = TERMINATOR

    def __bytes__(self):
        """Returns the packet as bytes with prepended length."""
        payload = bytes(self.id)
        payload += bytes(self.type)
        payload += self.payload
        payload += self.terminator
        size = bytes(LittleEndianSignedInt32(len(payload)))
        return size + payload

    @classmethod
    async def aread(cls, reader: StreamReader) -> Packet:
        """Reads a packet from an asynchronous file-like object."""
        size = await LittleEndianSignedInt32.aread(reader)
        id_ = await LittleEndianSignedInt32.aread(reader)
        type_ = await Type.aread(reader)
        payload = await reader.read(size - 10)
        terminator = await reader.read(2)

        if terminator != TERMINATOR:
            LOGGER.warning('Unexpected terminator: %s', terminator)

        return cls(id_, type_, payload, terminator)

    @classmethod
    def read(cls, file: IO) -> Packet:
        """Reads a packet from a file-like object."""
        size = LittleEndianSignedInt32.read(file)
        id_ = LittleEndianSignedInt32.read(file)
        type_ = Type.read(file)
        payload = file.read(size - 10)
        terminator = file.read(2)

        if terminator != TERMINATOR:
            LOGGER.warning('Unexpected terminator: %s', terminator)

        return cls(id_, type_, payload, terminator)

    @classmethod
    def make_command(cls, *args: str, encoding: str = 'utf-8') -> Packet:
        """Creates a command packet."""
        return cls(
            random_request_id(), Type.SERVERDATA_EXECCOMMAND,
            b' '.join(map(partial(str.encode, encoding=encoding), args))
        )

    @classmethod
    def make_login(cls, passwd: str, *, encoding: str = 'utf-8') -> Packet:
        """Creates a login packet."""
        return cls(
            random_request_id(), Type.SERVERDATA_AUTH, passwd.encode(encoding)
        )



"""Testing of the proto module."""

from functools import partial
from io import BytesIO
from random import randint
from unittest import TestCase

from rcon.proto import LittleEndianSignedInt32, Packet, Type, random_request_id


class TestRandomRequestId(TestCase):
    """Tests the random_request_id function."""

    def _test_type(self, request_id: LittleEndianSignedInt32):
        """Tests the type of a request id."""
        self.assertIsInstance(request_id, LittleEndianSignedInt32)

    def _test_value(self, request_id: LittleEndianSignedInt32):
        """Tests the value of a request id."""
        self.assertTrue(0 <= request_id <= LittleEndianSignedInt32.MAX)

    def _test_request_id(self, request_id: LittleEndianSignedInt32):
        """Tests a request id."""
        self._test_type(request_id)
        self._test_value(request_id)

    def test_random_request_ids(self):
        """Tests for valid values."""
        for _ in range(1000):
            self._test_request_id(random_request_id())


class TestLittleEndianSignedInt32(TestCase):
    """Tests the LittleEndianSignedInt32 type."""

    def test_min(self):
        """Tests the minimum value."""
        self.assertEqual(LittleEndianSignedInt32(LittleEndianSignedInt32.MIN),
                         LittleEndianSignedInt32.MIN)

    def test_max(self):
        """Tests the maximum value."""
        self.assertEqual(LittleEndianSignedInt32(LittleEndianSignedInt32.MAX),
                         LittleEndianSignedInt32.MAX)

    def test_below_min(self):
        """Tests a value below the minimum value."""
        self.assertRaises(
            ValueError,
            partial(LittleEndianSignedInt32, LittleEndianSignedInt32.MIN - 1)
        )

    def test_above_max(self):
        """Tests a value above the maximum value."""
        self.assertRaises(
            ValueError,
            partial(LittleEndianSignedInt32, LittleEndianSignedInt32.MAX + 1)
        )

    def test_random(self):
        """Tests random LittleEndianSignedInt32 values."""
        for _ in range(1000):
            random = randint(LittleEndianSignedInt32.MIN,
                             LittleEndianSignedInt32.MAX + 1)
            self.assertEqual(LittleEndianSignedInt32(random), random)


class TestType(TestCase):
    """Tests the Type enum."""

    def test_serverdata_auth_value(self):
        """Tests the SERVERDATA_AUTH value."""
        self.assertEqual(Type.SERVERDATA_AUTH.value, 3)
        self.assertEqual(int(Type.SERVERDATA_AUTH), 3)

    def test_serverdata_auth_bytes(self):
        """Tests the SERVERDATA_AUTH bytes."""
        self.assertEqual(bytes(Type.SERVERDATA_AUTH),
                         (3).to_bytes(4, 'little'))

    def test_serverdata_auth_read(self):
        """Tests reading of SERVERDATA_AUTH."""
        self.assertIs(Type.read(BytesIO((3).to_bytes(4, 'little'))),
                      Type.SERVERDATA_AUTH)

    def test_serverdata_auth_response_value(self):
        """Tests the SERVERDATA_AUTH_RESPONSE value."""
        self.assertEqual(Type.SERVERDATA_AUTH_RESPONSE.value, 2)
        self.assertEqual(int(Type.SERVERDATA_AUTH_RESPONSE), 2)

    def test_serverdata_auth_response_bytes(self):
        """Tests the SERVERDATA_AUTH_RESPONSE bytes."""
        self.assertEqual(bytes(Type.SERVERDATA_AUTH_RESPONSE),
                         (2).to_bytes(4, 'little'))

    def test_serverdata_auth_response_read(self):
        """Tests the reading of SERVERDATA_AUTH_RESPONSE."""
        self.assertIs(Type.read(BytesIO((2).to_bytes(4, 'little'))),
                      Type.SERVERDATA_AUTH_RESPONSE)

    def test_serverdata_execcommand_value(self):
        """Tests the SERVERDATA_EXECCOMMAND value."""
        self.assertEqual(Type.SERVERDATA_EXECCOMMAND.value, 2)
        self.assertEqual(int(Type.SERVERDATA_EXECCOMMAND), 2)

    def test_serverdata_execcommand_bytes(self):
        """Tests the SERVERDATA_EXECCOMMAND bytes."""
        self.assertEqual(bytes(Type.SERVERDATA_EXECCOMMAND),
                         (2).to_bytes(4, 'little'))

    def test_serverdata_execcommand_read(self):
        """Tests the reading of SERVERDATA_EXECCOMMAND."""
        self.assertIs(Type.read(BytesIO((2).to_bytes(4, 'little'))),
                      Type.SERVERDATA_EXECCOMMAND)

    def test_serverdata_response_value_value(self):
        """Tests the SERVERDATA_RESPONSE_VALUE value."""
        self.assertEqual(Type.SERVERDATA_RESPONSE_VALUE.value, 0)
        self.assertEqual(int(Type.SERVERDATA_RESPONSE_VALUE), 0)

    def test_serverdata_response_value_bytes(self):
        """Tests the SERVERDATA_RESPONSE_VALUE bytes."""
        self.assertEqual(bytes(Type.SERVERDATA_RESPONSE_VALUE),
                         (0).to_bytes(4, 'little'))

    def test_serverdata_response_value_read(self):
        """Tests the reading SERVERDATA_RESPONSE_VALUE."""
        self.assertIs(Type.read(BytesIO((0).to_bytes(4, 'little'))),
                      Type.SERVERDATA_RESPONSE_VALUE)


class TestPacket(TestCase):
    """Tests the Packet named tuple."""

    def setUp(self):
        """Creates a packet."""
        self.packet = Packet(
            random_request_id(),
            Type.SERVERDATA_EXECCOMMAND,
            'Lorem ipsum sit amet...'.encode()
        )

    def test_bytes_rw(self):
        """Tests recovering from bytes."""
        self.assertEqual(Packet.read(BytesIO(bytes(self.packet))), self.packet)



"""Tests the configuration file parsing."""

from itertools import product
from random import shuffle
from string import printable
from typing import Iterator, Tuple
from unittest import TestCase

from rcon.config import Config


def random_passwd() -> str:
    """Generates a random password containing all printable characters."""

    chars = list(printable)
    shuffle(chars)
    return ''.join(chars)


class TestConfig(TestCase):
    """Test the named tuple Config."""

    def setUp(self):
        """Sets up test and target data."""
        self.hosts = [
            'subsubdomain.subdomain.example.com',
            'locahost',
            '127.0.0.1'
        ]
        self.ports = range(65_536)

    @property
    def _sockets(self) -> Iterator[Tuple[str, int]]:
        """Yields (host, port) tuples."""
        return product(self.hosts, self.ports)

    def _test_from_string_with_password(self, host, port):
        """Tests the Config.from_string() method with a password."""
        passwd = random_passwd()
        config = Config.from_string(f'{passwd}@{host}:{port}')
        self.assertEqual(config.host, host)
        self.assertEqual(config.port, port)
        self.assertEqual(config.passwd, passwd)

    def _test_from_string_without_password(self, host, port):
        """Tests the Config.from_string() method without a password."""
        config = Config.from_string(f'{host}:{port}')
        self.assertEqual(config.host, host)
        self.assertEqual(config.port, port)
        self.assertIsNone(config.passwd)

    def test_from_string(self):
        """Tests the Config.from_string() method."""
        for host, port in self._sockets:
            self._test_from_string_with_password(host, port)
            self._test_from_string_without_password(host, port)



"""Asynchronous RCON."""

from asyncio import StreamReader, StreamWriter, open_connection

from rcon.exceptions import RequestIdMismatch, WrongPassword
from rcon.proto import Packet, Type


__all__ = ['rcon']


async def close(writer: StreamWriter) -> None:
    """Close socket asynchronously."""

    writer.close()
    await writer.wait_closed()


async def communicate(
        reader: StreamReader,
        writer: StreamWriter,
        packet: Packet
) -> Packet:
    """Asynchronous requests."""

    writer.write(bytes(packet))
    await writer.drain()
    return await Packet.aread(reader)


async def rcon(
        command: str,
        *arguments: str,
        host: str,
        port: int,
        passwd: str,
        encoding: str = 'utf-8'
) -> str:
    """Runs a command asynchronously."""

    reader, writer = await open_connection(host, port)
    login = Packet.make_login(passwd, encoding=encoding)
    response = await communicate(reader, writer, login)

    # Wait for SERVERDATA_AUTH_RESPONSE according to:
    # https://developer.valvesoftware.com/wiki/Source_RCON_Protocol
    while response.type != Type.SERVERDATA_AUTH_RESPONSE:
        response = await Packet.aread(reader)

    if response.id == -1:
        await close(writer)
        raise WrongPassword()

    request = Packet.make_command(command, *arguments, encoding=encoding)
    response = await communicate(reader, writer, request)
    await close(writer)

    if response.id != request.id:
        raise RequestIdMismatch(request.id, response.id)

    return response.payload.decode(encoding)



"""RCON client CLI."""

from argparse import ArgumentParser, Namespace
from logging import DEBUG, INFO, basicConfig, getLogger
from pathlib import Path

from rcon.client import Client
from rcon.config import CONFIG_FILES, LOG_FORMAT, from_args
from rcon.errorhandler import ErrorHandler


__all__ = ['main']


LOGGER = getLogger('rconclt')


def get_args() -> Namespace:
    """Parses and returns the CLI arguments."""

    parser = ArgumentParser(description='A Minecraft RCON client.')
    parser.add_argument('server', help='the server to connect to')
    parser.add_argument(
        '-c', '--config', type=Path, metavar='file', default=CONFIG_FILES,
        help='the configuration file'
    )
    parser.add_argument(
        '-d', '--debug', action='store_true',
        help='print additional debug information'
    )
    parser.add_argument(
        '-t', '--timeout', type=float, metavar='seconds',
        help='connection timeout in seconds'
    )
    parser.add_argument('command', help='command to execute on the server')
    parser.add_argument(
        'argument', nargs='*', default=(), help='arguments for the command'
    )
    return parser.parse_args()


def run() -> None:
    """Runs the RCON client."""

    args = get_args()
    basicConfig(format=LOG_FORMAT, level=DEBUG if args.debug else INFO)
    host, port, passwd = from_args(args)

    with Client(host, port, timeout=args.timeout) as client:
        client.login(passwd)
        text = client.run(args.command, *args.argument)

    print(text, flush=True)


def main() -> int:
    """Runs the main script with exceptions handled."""

    with ErrorHandler(LOGGER) as handler:
        run()

    return handler.exit_code



"""RCON client library."""

from rcon.async_rcon import rcon
from rcon.client import Client
from rcon.exceptions import RequestIdMismatch, WrongPassword


__all__ = ['RequestIdMismatch', 'WrongPassword', 'Client', 'rcon']



"""An interactive RCON shell."""

from argparse import ArgumentParser, Namespace
from logging import INFO, basicConfig, getLogger
from pathlib import Path

from rcon.config import CONFIG_FILES, LOG_FORMAT, from_args
from rcon.console import PROMPT, rconcmd
from rcon.errorhandler import ErrorHandler
from rcon.readline import CommandHistory


__all__ = ['get_args', 'main']


LOGGER = getLogger('rconshell')


def get_args() -> Namespace:
    """Parses and returns the CLI arguments."""

    parser = ArgumentParser(description='An interactive RCON shell.')
    parser.add_argument('server', nargs='?', help='the server to connect to')
    parser.add_argument(
        '-c', '--config', type=Path, metavar='file', default=CONFIG_FILES,
        help='the configuration file'
    )
    parser.add_argument(
        '-p', '--prompt', default=PROMPT, metavar='PS1',
        help='the shell prompt'
    )
    return parser.parse_args()


def run() -> None:
    """Runs the RCON shell."""

    args = get_args()
    basicConfig(level=INFO, format=LOG_FORMAT)

    if args.server:
        host, port, passwd = from_args(args)
    else:
        host = port = passwd = None

    with CommandHistory(LOGGER):
        rconcmd(host, port, passwd, prompt=args.prompt)


def main() -> int:
    """Runs the main script with exceptions handled."""

    with ErrorHandler(LOGGER) as handler:
        run()

    return handler.exit_code



"""RCON exceptions."""

__all__ = [
    'ConfigReadError',
    'RequestIdMismatch',
    'UserAbort',
    'WrongPassword'
]


class ConfigReadError(Exception):
    """Indicates an error while reading the configuration."""


class RequestIdMismatch(Exception):
    """Indicates a mismatch in request IDs."""

    def __init__(self, sent: int, received: int):
        """Sets the IDs that have been sent and received."""
        super().__init__()
        self.sent = sent
        self.received = received


class UserAbort(Exception):
    """Indicates that a required action has been aborted by the user."""


class WrongPassword(Exception):
    """Indicates a wrong password."""



"""RCON server configuration."""

from __future__ import annotations
from argparse import Namespace
from configparser import ConfigParser, SectionProxy
from getpass import getpass
from logging import getLogger
from os import getenv, name
from pathlib import Path
from typing import Iterable, NamedTuple, Optional, Union

from rcon.exceptions import ConfigReadError, UserAbort


__all__ = ['CONFIG_FILES', 'LOG_FORMAT', 'SERVERS', 'Config', 'from_args']


CONFIG = ConfigParser()

if name == 'posix':
    CONFIG_FILES = (
        Path('/etc/rcon.conf'),
        Path('/usr/local/etc/rcon.conf'),
        Path.home().joinpath('.rcon.conf')
    )
elif name == 'nt':
    CONFIG_FILES = (
        Path(getenv('LOCALAPPDATA')).joinpath('rcon.conf'),
        Path.home().joinpath('.rcon.conf')
    )
else:
    raise NotImplementedError(f'Unsupported operating system: {name}')

LOG_FORMAT = '[%(levelname)s] %(name)s: %(message)s'
LOGGER = getLogger('RCON Config')
SERVERS = {}


class Config(NamedTuple):
    """Represents server configuration."""

    host: str
    port: int
    passwd: Optional[str] = None

    @classmethod
    def from_string(cls, string: str) -> Config:
        """Reads the credentials from the given string."""
        try:
            host, port = string.rsplit(':', maxsplit=1)
        except ValueError:
            raise ValueError(f'Invalid socket: {string}.') from None

        port = int(port)

        try:
            passwd, host = host.rsplit('@', maxsplit=1)
        except ValueError:
            passwd = None

        return cls(host, port, passwd)

    @classmethod
    def from_config_section(cls, section: SectionProxy) -> Config:
        """Creates a credentials tuple from
        the respective config section.
        """
        host = section['host']
        port = section.getint('port')
        passwd = section.get('passwd')
        return cls(host, port, passwd)


def load(config_files: Union[Path, Iterable[Path]] = CONFIG_FILES) -> None:
    """Reads the configuration files and populates SERVERS."""

    SERVERS.clear()
    CONFIG.read(config_files)

    for section in CONFIG.sections():
        SERVERS[section] = Config.from_config_section(CONFIG[section])


def from_args(args: Namespace) -> Config:
    """Get the credentials for a server from the respective arguments."""

    try:
        host, port, passwd = Config.from_string(args.server)
    except ValueError:
        load(args.config)

        try:
            host, port, passwd = SERVERS[args.server]
        except KeyError:
            LOGGER.error('No such server: %s.', args.server)
            raise ConfigReadError() from None

    if passwd is None:
        try:
            passwd = getpass('Password: ')
        except (KeyboardInterrupt, EOFError):
            print()
            LOGGER.error('Aborted by user.')
            raise UserAbort() from None

    return Config(host, port, passwd)



#! /usr/bin/env python
"""Installation script."""

from setuptools import setup

setup(
    name='rcon',
    use_scm_version=True,
    setup_requires=['setuptools_scm'],
    author='Richard Neumann',
    author_email='mail@richard-neumann.de',
    python_requires='>=3.8',
    packages=['rcon'],
    extras_require={'GUI':  ['pygobject', 'pygtk']},
    entry_points={
        'console_scripts': [
            'rcongui = rcon.gui:main',
            'rconclt = rcon.rconclt:main',
            'rconshell = rcon.rconshell:main',
        ],
    },
    url='https://github.com/conqp/rcon',
    license='GPLv3',
    description='An RCON client library.',
    long_description=open('README.md').read(),
    long_description_content_type="text/markdown",
    keywords='python rcon client'
)



"""Synchronous client."""

from socket import socket
from typing import Optional

from rcon.exceptions import RequestIdMismatch, WrongPassword
from rcon.proto import Packet, Type


__all__ = ['Client']


class Client:
    """An RCON client."""

    __slots__ = ('_socket', 'host', 'port', 'passwd')

    def __init__(
            self, host: str, port: int, *,
            timeout: Optional[float] = None,
            passwd: Optional[str] = None
    ):
        """Initializes the base client with the SOCK_STREAM socket type."""
        self._socket = socket()
        self.host = host
        self.port = port
        self.timeout = timeout
        self.passwd = passwd

    def __enter__(self):
        """Attempts an auto-login if a password is set."""
        self._socket.__enter__()
        self.connect(login=True)
        return self

    def __exit__(self, typ, value, traceback):
        """Delegates to the underlying socket's exit method."""
        return self._socket.__exit__(typ, value, traceback)

    @property
    def timeout(self) -> float:
        """Returns the socket timeout."""
        return self._socket.gettimeout()

    @timeout.setter
    def timeout(self, timeout: float):
        """Sets the socket timeout."""
        self._socket.settimeout(timeout)

    def connect(self, login: bool = False) -> None:
        """Connects the socket and attempts a
        login if wanted and a password is set.
        """
        self._socket.connect((self.host, self.port))

        if login and self.passwd is not None:
            self.login(self.passwd)

    def close(self) -> None:
        """Closes the socket connection."""
        self._socket.close()

    def communicate(self, packet: Packet) -> Packet:
        """Sends and receives a packet."""
        with self._socket.makefile('wb') as file:
            file.write(bytes(packet))

        return self.read()

    def read(self) -> Packet:
        """Reads a packet."""
        with self._socket.makefile('rb') as file:
            return Packet.read(file)

    def login(self, passwd: str, *, encoding: str = 'utf-8') -> bool:
        """Performs a login."""
        request = Packet.make_login(passwd, encoding=encoding)
        response = self.communicate(request)

        # Wait for SERVERDATA_AUTH_RESPONSE according to:
        # https://developer.valvesoftware.com/wiki/Source_RCON_Protocol
        while response.type != Type.SERVERDATA_AUTH_RESPONSE:
            response = self.read()

        if response.id == -1:
            raise WrongPassword()

        return True

    def run(self, command: str, *args: str, encoding: str = 'utf-8') -> str:
        """Runs a command."""
        request = Packet.make_command(command, *args, encoding=encoding)
        response = self.communicate(request)

        if response.id != request.id:
            raise RequestIdMismatch(request.id, response.id)

        return response.payload.decode(encoding)
